(** Main file
@author Sarah Winkler
@since  2010/12/30 *)

(*** OPENS **************************************************************)
open Util;;

(*** SUBMODULES *********************************************************)
module E = Either;;
module C = Completion;;
module W = World;;
module Label = U.Label;;
module Term = U.Term;;
module RMonad = U.Monad;;
module WMonad = World.Monad;;
module TrsParser = U.MyTrsParser;;
module TrsLexer = U.MyTrsLexer;;
module TrsSyntax = U.MyTrsSyntax;;
module Problem = U.MyProblem;;
module St = Statistics;;
module NS = NodeState;;

(*** OPENS (2) **********************************************************)
open WMonad;;
open Completion;;

(*** EXCEPTIONS *********************************************************)
exception Help;;

(*** GLOBALS ************************************************************)
let flags = ref C.default_options;;
let set_timeout f = flags := {!flags with timeout = f};;
let set_ttimeout f = flags := {!flags with ttimeout = f};;
let set_kill_processes r = flags := {!flags with kill_processes = r};;
let set_output_trs () = flags := {!flags with output_trs = true};;
let set_output_stats () = flags := {!flags with output_statistics = true};;
let set_theory f = flags := {!flags with s_theory_file = f};;
let set_verbosity v = flags := {!flags with verbosity = v};;

let set_cpc s = 
 let cpc = 
  if (String.compare s "none") == 0 then C.NoCPC
  else if (String.compare s "connected") == 0 then C.Connected
  else if (String.compare s "all") == 0 then C.Mixed
  else if (String.compare s "sred") == 0 then C.SReducibility
  else C.Prime
 in flags := {!flags with cpc = cpc}
;;

let set_internal s = 
 flags := {!flags with check_externally = false; termination_strategy=s};;
let set_ttool s = 
 flags := {!flags with tool=s; check_externally = true};;

let set_selection s =
 if (String.compare s "sum") == 0 then
  flags := {!flags with
    selection_strategy = SelectionStrategy.japan_sum;
    strategy_macro = Some(SelectionStrategy.JapanSum)}
 else if (String.compare s "max") == 0 then
  flags := {!flags with
    selection_strategy = SelectionStrategy.japan_max;
    node_size = C.Max;
    strategy_macro = Some(SelectionStrategy.JapanMax)}
 else
  (flags := {!flags with strategy_macro = None};
   if (String.compare s "slothrop") == 0 then
    flags := {!flags with selection_strategy = SelectionStrategy.slothrop}
  else
   let strategy = SelectionStrategyAux.of_string s in
   flags := {!flags with selection_strategy = strategy}
  )
;;

(*                         DEFINE SWITCHES                              *)
let sws = [
 ("-cp", Arg.String set_cpc,
  "critical pair criterion: sred, prime, connected, all or none");
 ("-ct", Arg.Unit set_output_trs,
  "output completed TRS");
 ("-it", Arg.String set_internal,
  "check termination internally with given strategy");
 ("-kp", Arg.Float set_kill_processes,
  "threshold to kill processes, > 1.0");
 ("-st", Arg.Unit set_output_stats,
  "output statistics");
 ("-ss", Arg.String set_selection,
  "selection strategy");
 ("-t", Arg.Float set_timeout,
  "global time limit");
 ("-T", Arg.Float set_ttimeout,
  "termination call time limit");
 ("-th", Arg.String set_theory,
  "theory S with respect to which completion performed");
 ("-tp", Arg.String set_ttool,
  "executable to call external AC-termination prover");
 ("-v", Arg.Int set_verbosity,
  "verbosity in {0, 1, 2}");

]

let sws = Arg.alignx 80 sws;;

let usage = "Usage: ./tonc [options] <file> [timeout]\n\nOptions:";;

(*** FUNCTIONS ***********************************************************)
let log0 = W.log 0;;
let log1 = W.log 1;;
let log2 = W.log 2;;

let configurate argv =
 if Array.length argv = 1 then (Arg.usagex sws usage; exit 0);
 let file_is_set = ref true in
 let push_file n =
  if !file_is_set then
   (file_is_set := false;
   flags := {!flags with filename = n})
  else
   (Arg.usagex sws usage; exit 0)
 in
 Arg.parse sws push_file usage;
;;

(*                             READ INPUT                               *)
let read_wst_file filename =
 let chin = Pervasives.open_in filename in
 let lbuf = Lexing.from_channel chin in
 let ts = TrsParser.trs TrsLexer.token lbuf in
 liftm RMonad.get >>= fun s ->
 let p, s = TrsSyntax.to_problem_with ts s in
 liftm (RMonad.set s) >>
 let trs = Problem.get_trs p in
 return trs
;;

let read_file filename =
 if Filename.check_suffix filename "tptp" then
  failwith "No TPTP yet"
 else
  read_wst_file filename >>= fun trs ->
  let eqs = Trsx.to_equations trs in
  W.set_axioms eqs >> return eqs
;;

(*                               OUTPUT                                 *)
let if_do b m = if b then m else return ()

let output (status, result) t =
 let problem_name = Filename.basename !flags.filename in
 St.print_basic status problem_name t;
 begin
 match Completion.mode (!flags),status,result with
  | C.Completion,C.Sat,Some p ->
   if_do !flags.output_proof (Control.print_proof status p) >>
   NS.project_r_closed(*_unprotected *) p >>= W.M.Trsx.readable >>= fun r ->
   if_do !flags.output_trs  (St.print_system r) >>
   let e0 = !flags.axioms in
   if_do !flags.output_statistics (St.print_success e0 p t)
  | C.Completion, C.Gup _, _
  | C.Completion, C.Unk, _
  | C.Completion, C.Tmo, _ ->
   if_do !flags.output_statistics (St.print_timeout t)
  | C.CompletenessCheck _, C.Suc,  _ ->
   log0 "Completeness and joinability of axioms verified."
  | C.CompletenessCheck _, C.Noc, _ ->
   log0 "Completeness and joinability of axioms not verified."
  | _ -> raise (C.GiveUp "Strange result state")
 end >> St.print_logs
;;

let set_theory eqs sg =
 if String.length (!flags.s_theory_file) <> 0 then (
  read_wst_file !flags.s_theory_file >>= fun s ->
  flags := {!flags with s_theory = s};
  return ())
 else (
  liftm (Theories.Find.largest eqs) >>= fun (_,s,n) ->
(*  Format.printf "Recognized %s\n%!" n;*)
  flags := {!flags with s_theory = s; s_theory_file = n};
  return ())
;;

let work () =
 read_file !flags.filename >>= fun eqs ->
 liftm RMonad.get >>= fun sg ->
 set_theory eqs sg >>= fun _ ->
 W.Monad.set (W.initial_context_with !flags) >>
 let tstart = Unix.gettimeofday () in
 Control.complete eqs >>= fun result ->
 let d = (Unix.gettimeofday () -. tstart) in 
 output result d >> return ""
;;

let run_monad () =
 let s = Theories.sigma in
 let w = W.initial_context_with !flags in
 let r = RMonad.run s (W.Monad.run w (work ())) in
 if E.is_right r then E.right r else failwith "No result from Monad.run"
;;

let output = function
 | None,_ -> Format.printf "Timeout\n"
 | Some s,t -> Format.printf "Total time: %f\n%s" (!flags.timeout-.t) s
;; 

let execute () =
 try
  configurate Sys.argv;
  let t = !flags.timeout in
  output (Process.run_timed (Process.Local t) run_monad ())
 with | Help -> Arg.usagex sws usage;
;;

execute ()
