(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module C = Complexity;;
module F = Format;;
module G = Graph;;
module L = List;;
module M = Monad;;
module Map = Map.Partial (Term);;
module P = Problem;;
module Fun = Function;;
module R = Rule;;
module S = Substitution;;

(*** TYPES ********************************************************************)
type flags = {help : bool ref; i : bool ref};;
type t = P.t * P.t;;

(*** GLOBALS ******************************************************************)
let code = "acdg";;
let name = "AC-EDG Processor";;
let keywords = ["dependency graph";"approximation";"AC-termination"];;
let flags = {help = ref false; i = ref false};;

let comment =
 "Remove all edges from the current DG that are not contained in the \
  AC-EDG (approximation of AC-DG based on AC-unification)."
;;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.")]
 in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,L.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = M.(>>=);;
let init _ = flags.help := false; flags.i := false;;

(* Destructors *)
let get_ip = fst;; (* input problem *)
let get_op = snd;; (* output problem *)

(* Processor *)
(* ets are terms in sharped equations except f(x,y) *)
let is_arc ets mr (x,y) =
 let t = R.rhs x and u = R.lhs y in
 (* (1) *)
 let ts = L.filter (Elogic.are_unifiable t) ets in
 if L.exists ((=) (Term.root u) <.> Term.root) ts then M.return true
 (* (2) *)
 else Aclogic.are_unifiable (Map.find t mr) u
;;

let generate ets mr ns =
 M.foldl (fun g n -> M.foldl (fun ms m ->
  M.lift (fun c -> if c then m::ms else ms) (is_arc ets mr (n,m))) [] ns >>=
  (M.return <.> flip (G.add n) g)) G.empty ns
;;

let filter_edges ets mr g =
 G.fold (fun n ms m -> m >>= fun g ->
  let add m ms c = M.return (if c then m::ms else ms) in
  M.foldl (fun ms m -> is_arc ets mr (n,m) >>= add m ms) [] ms >>=
  (M.return <.> flip (G.add n) g)) g (M.return G.empty)
;;

let nontrivial = function
 | Term.Fun(f, [x;y]) -> M.is_theory Label.AC f >>= fun ac -> 
   M.return (not (ac && (List.for_all Term.is_var [x;y])))
 | _ -> M.return true
;;

let solve fs p =
 let configurate s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 if P.is_edp p then (* if equational DP problem *)
  let trs = P.get_trs p and dps = P.get_dps p and g = P.get_dg p in
  M.filter nontrivial (Trs.terms (P.get_eqssharp p)) >>= fun ets ->
  let (ls,rs) = match g with
   | P.Complete -> (Trs.lhs dps,Trs.rhs dps)
   | P.Partial g -> (L.map R.lhs (G.in_nodes g),L.map R.rhs (G.out_nodes g))
  in
  let rs = L.unique_hash rs in
  let add f m t = f t >>= (M.return <.> flip (Map.add t) m) in
  M.foldl (add (flip Trs.ac_rencap trs)) Map.empty rs >>= fun mr ->
  (* compute arcs *)
  match g with
   | P.Complete ->
    generate ets mr (Trs.to_list dps) >>= fun g ->
    let n = Int.square (Trs.size dps) and m = G.size_edges g in
    M.return (if m < n then Some (p,P.set_dg (P.Partial g) p) else None)
   | P.Partial g ->
    let g = G.restrict (Trs.to_list dps) g in
    filter_edges ets mr g >>= fun h ->
    let n = G.size_edges g and m = G.size_edges h in
    M.return (if m < n then Some (p,P.set_dg (P.Partial h) p) else None)
 else M.return None
;;

(* Complexity Bounds *)
let complexity c _ = C.mul c C.other;;

(* Compare Functions *)
let equal p q = P.equal (get_ip p) (get_ip q) 

(* Printers *)
let fprintf fs fmt p =
 F.fprintf fmt "@[<1>%s:@\n" name;
 P.fprintfm ~g:true fmt (get_op p) >>= fun _ ->
 F.fprintf fmt "@\n"; L.hd fs fmt >>= fun _ ->
 M.return (F.fprintf fmt "@]")
;;

let fprintfx fs fmt p =
 F.fprintf fmt "@{<dependencyGraph>";
 P.fprintfx ~g:true fmt (get_op p) >>= fun _ ->
 L.hd fs fmt >>= fun _ -> M.return (F.fprintf fmt "@}")
;;
