(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;
open Logic.Operators;;

(*** MODULES ******************************************************************)
module AF = Filtering;;
module C = Coefficient;;
module Co = Complexity;;
module F = Format;;
module Fun = Function;;
module H = Hashtbl;;
module M = Rewritingx.Monad;;
module Number = Logic.Number;;
module P = Problem;;
module Pos = Position;;
module Prec = Precedence;;
module Sig = Signature;;
module Var = Variable;;

module Status = struct
 (*** INCLUDES ****************************************************************)
 include Index.Make (Fun) (Bool);;

 (*** FUNCTIONS ***************************************************************)
 let (>>=) = M.(>>=);;

 let is_lex s f = not (find f s);;
 let is_mul s f = find f s;;

 let fprintf fmt p =
  let rec fprintf fmt = function
   | [] -> ()
   | (f,b) :: p ->
    if b then F.fprintf fmt "%a:mul@ @ %a" Fun.fprintf f fprintf p
    else F.fprintf fmt "%a:lex@ @ %a" Fun.fprintf f fprintf p
 in
 F.fprintf fmt "@[%a@]" fprintf (to_list p)
;;

let fprintfm fmt p =
 let rec fprintfm = function
  | [] -> M.return ()
  | (f,b) :: p ->
   M.fprintf_fun fmt f >>= fun _ ->
   if b then F.fprintf fmt ":lex @ " else F.fprintf fmt ":mul @ ";
   fprintfm p
 in
 M.filter (fun (f,_) -> M.is_theory Label.AC f >>= (M.return <.> not)) (to_list p) >>=
 M.filter (fun (f,_) -> M.find_ari f >>= fun a -> M.return (a > 1)) >>=
 (F.fprintf fmt "@["; fprintfm) >>= fun _ ->
 M.return (F.fprintf fmt "@]")
;;

let fprintfx fmt p s =
  let p = Prec.to_list p in
  F.fprintf fmt "@{<statusPrecedence>";
  M.iter (fun(f,b) ->
    F.fprintf fmt "@{<statusPrecedenceEntry>";
    M.fprintfx_fun fmt f >>= fun _ ->
    M.find_ari f         >>= fun a ->
    F.fprintf fmt "@{<arity>%i@}" a;
    F.fprintf fmt "@{<precedence>%i@}" (List.assoc f p);
    F.fprintf fmt "@{<%s>@}" (if b then "mul" else "lex");
    M.return(F.fprintf fmt "@}")
  ) (to_list s) >>= fun _ ->
  M.return(F.fprintf fmt "@}")
;;

end;;




(*** TYPES ********************************************************************)

type context = {
 solver             : Logic.solver;
 prec_num           : int;
 state              : Signature.t;
 precs              : (Fun.t, Logic.a) H.t;
 gt_encodings       : (Rule.t, Logic.p) H.t;
 (* encoding statuses of function symbols; true is multiset *)
 stats              : (Fun.t, Logic.p) H.t;
 (* multiset cover: if (s->t, j) has value i then both s,t are rooted by f and 
    in multiset comparison of args(s), args(t) term t_j is covered by s_i *)
 mcov_a_encodings   : (Term.t list * Term.t list * int, Logic.a) H.t;
 (* multiset comparison formula caching *)
 mcov_encodings     : (Term.t list*Term.t list, Logic.p*Logic.a list(**Logic.p list*)) H.t;
};;

type flags = {
 direct : bool ref;
 help : bool ref;
(* quasi : bool ref;*)
 sat : bool ref;
};;

type t = {
 input : P.t;
 output : P.t;
 precedence : Prec.t;
 status : Status.t;
};;

(*** GLOBALS ******************************************************************)
let code = "acrpo";;
let name = "AC-RPO Processor";;
let comment = "Applies AC-Recursive Path Order."

let keywords =
 ["AC";"recursive path order";"simplification order";"termination"]
;;

let flags = {
 direct = ref false;
 help = ref false;
(* quasi = ref false;*)
 sat = ref false;
};;

let spec =
 let spec = [
  ("-direct",Arg.Set flags.direct,
   "Try to finish termination proof.");
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
 (* ("-quasi",Arg.Set flags.quasi,"Allows quasi-precedences (currently not \
   supported together with -dp flag).");*)
  ("-sat",Arg.Set flags.sat,"Uses SAT backend.");
  ("-smt",Arg.Clear flags.sat,"Uses SMT backend (default).")]
 in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** MODULES (part 2) *********************************************************)
module Statex = struct type t = context end;;
module Made = Util.Monad.Transformer.State (Statex) (Logic.Monad);;
open Made;;

(*** FUNCTIONS ****************************************************************)
let init _ =
 flags.direct := false;
 flags.help := false;
(* flags.quasi := false;*)
 flags.sat := false;
;;

(* Constructors and Destructors *)
let make input output precedence st = {
 input = input;
 output = output;
 precedence = precedence;
 status = st;
};;

let get_ip p = p.input;;
let get_op p = p.output;;

(* Complexity Bounds *)
let complexity c _ = Co.mul c Co.other;;

(* Compare Functions *)
let equal p q =
 P.equal p.input q.input && P.equal p.output q.output
;;

(* Printers *)
let (>>=) = M.(>>=);;

let fprintf fs fmt p  = 
 F.fprintf fmt "@[<1>%s:" name;
 F.fprintf fmt "@]@\n@[<1>precedence:@\n";
 Prec.fprintfm fmt p.precedence >>= fun _ ->
 F.fprintf fmt "@]@\n@[<1>status:@\n";
 Status.fprintfm fmt p.status >>= fun _ ->
 F.fprintf fmt "@]@\n@[<1>problem:@\n";
 P.fprintfm fmt p.output >>= fun _ ->
 F.fprintf fmt "@]@\n"; List.hd fs fmt >>= fun _ ->
 Monad.return (F.fprintf fmt "@]")
;;

let fprintfx_redpair fmt p =
  F.fprintf fmt "@{<redPair>@{<pathOrder>";
  Status.fprintfx fmt p.precedence p.status >>= fun _ ->
  M.return(F.fprintf fmt "@}@}")
;;

let fprintfx fs fmt p = 
 if (Problem.is_dp p.input) || (Problem.is_edp p.input) then (
  let tag = "redPairProc" in
  F.fprintf fmt "@{<%s>" tag;
  fprintfx_redpair fmt p         >>= fun _ ->
  Problem.fprintfx fmt p.output  >>= fun _ ->
  List.hd fs fmt                 >>= fun _ ->
  Monad.return(F.fprintf fmt "@}");
) else if (Problem.is_sp p.input) || (Problem.is_ep p.input)  then (
  failwith "AC-RPO full termination not supported!"
) else (
  failwith "for this problem AC-RPO is not supported!"
)
;;

(* general functions *)
let (>>=) = Made.(>>=);;

(* functions lifted from Logic into Made *)
let fresh_arith_spec spec = liftm (Logic.fresh_arith spec);;
let fresh_bool = get >>= fun s -> liftm Logic.fresh_bool;;

let eval_a a ass = a >>= fun a -> liftm (Logic.eval_a a ass);;
let eval_p p ass = p >>= fun p -> liftm (Logic.eval_p p ass);;

let ($>=$) a b = lift2 (<>=>) a b;;
let ($=$) a b = lift2 (<=>) a b;;
let ($>$) a b = lift2 (<>>) a b;;
let ($->$) a b = lift2 (<->>) a b;;
let ($|$) a b = lift2 (<|>) a b;;
let ($&$) a b = lift2 (<&>) a b;;
let ($*$) a b = lift2 (<*>) a b;;
let ($+$) a b = lift2 (<+>) a b;;
let ($-$) a b = lift2 (<->) a b;;
let mnot a = lift (~!) a;;
let zero = Logic.zero;;
let one = Logic.one;;
let constant = Logic.constant <.> Number.of_int;; 
let sum f = List.foldl (fun s x -> f x <+> s) Logic.zero;;
let big_xor ps = sum (fun p -> p <?> one <:> zero) ps <=> one;;

let map_op op f ls = sequence (List.map f ls) >>= (return <.> op);;
let mapi_op op f ls = sequence (List.mapi f ls) >>= (return <.> op);;
let gen_op op f n = sequence (List.gen f n) >>= (return <.> op);;
let map_and f = map_op Logic.big_and f;;
let mapi_and f = mapi_op Logic.big_and f;;
let map_or f = map_op Logic.big_or f;;

(* actual content starts here *)
let context state fs =
 let solver = if !(flags.sat) then Logic.MiniSat else Logic.Yices in
 {
  solver            = solver;
  prec_num          = max 0 (List.length fs - 1);
  precs             = H.create 512;
  state             = state;
  gt_encodings      = H.create 512;
  stats             = H.create 512;
  mcov_a_encodings  = H.create 512;
(*  mcov_p_encodings  = H.create 512;*)
  mcov_encodings    = H.create 512;
 }
;;

(* administrative functions *)
let cache_m tbl f k = 
 if H.mem tbl k then return (H.find tbl k)
 else (f k >>= fun v -> (H.add tbl k v; return v))
;;

(* encoding starts here *)
(* caching variables *)
let prec f = get >>= fun c -> 
 let arith = Logic.nat c.prec_num in
 cache_m c.precs (const (fresh_arith_spec arith)) f
;;

let prec_gt f g = prec f $>$ prec g;;
let prec_eq f g = return (if f = g then Logic.top else Logic.bot);; (* not quasi *)
let is_mul f = get >>= fun c -> cache_m c.stats (const fresh_bool) f;;
let is_lex = mnot <.> is_mul;;
let is_ac f = get >>= fun c -> return (Sig.is_theory Label.AC f c.state);;
let set_ac f =
 get >>= fun c -> let (f, s) = Sig.set_theory f Label.AC c.state in
 Made.set {c with state = s} >>
 return f
;;


(* *** *)
let find_term_with f =
 let rooted_f = function Term.Fun (g,_) when f=g -> true | _ -> false in
 let rec find ys = function
  | [] -> None
  | x::xs when rooted_f x -> Some(x, ys@xs)
  | x::xs -> find (x::ys) xs
 in find []
;;

(* sorts list of terms such that constants come before compound
   terms where the root has positive arity come before variables *)
let rec lex = function
  | [],[] -> 0
  | [],_ -> -1
  |_, [] -> 1
  | x::xs,y::ys -> let c = my_compare x y in if c=0 then lex (xs,ys) else c
and my_compare t t' =
 match t, t' with
  | Term.Var x, Term.Var y -> Var.compare x y
  | Term.Fun _, Term.Var _ -> -1
  | Term.Var _, Term.Fun _ -> 1
  | Term.Fun(_,[]), Term.Fun(_,t::ts) -> -1
  | Term.Fun(_,t::ts), Term.Fun(_,[]) -> 1
  | Term.Fun(f,fs), Term.Fun(g,gs) when f=g -> lex (fs,gs)
  | Term.Fun(f,_), Term.Fun(g,_) -> Fun.compare f g
;;

(* recursively flatten term with respect to all ac symbols *)
let rec flatten = function
 | Term.Fun(f, ts) ->
  is_ac f >>= fun f_is_ac -> (
  if f_is_ac then
   Made.map flatten ts >>= fun ts' ->
   match find_term_with f ts' with
    | None -> Made.return (Term.Fun(f, List.sort my_compare ts'))
    | Some (ti,ts'') -> flatten (Term.Fun(f, (ts''@(Term.args ti))))
  else
   Made.map flatten ts >>= fun us -> Made.return (Term.Fun(f,us)))
 | v -> Made.return v
;;

let rec term_is_embedded s t = match (s,t) with
  | Term.Var x, Term.Var y -> x = y
  | Term.Fun (f,ss), Term.Fun (g,ts) ->
   (f = g 
    && List.length ss = List.length ts 
    && List.for_all2 term_is_embedded ss ts) || List.exists (term_is_embedded s) ts
  | _, Term.Fun (f,ts) -> List.exists (term_is_embedded s) ts
  | _, _ -> false
;;

let is_embedded = Util.uncurry (flip term_is_embedded) <.> Rule.to_terms;;
let is_proper_embedded r =
  List.exists (term_is_embedded (Rule.rhs r)) (Term.subterms (Rule.lhs r))
 ;;
(* *** *)

(* multiset cover arith variable caching, where key=(ss,tt,j) *)
let mcov_a_var key =
 let m = List.length (Triple.fst key) -1 in
 get >>= fun c -> let arith = Logic.nat m in
 cache_m c.mcov_a_encodings (const (fresh_arith_spec arith)) key
;;

(* encoding *)
let rec lex gt eq ts ss = match ts,ss with
 | _::_,[] -> return Logic.top
 | [], _ -> return Logic.bot
 | ti::ts,si::ss -> 
  gt (Rule.of_terms ti si) $|$ (eq (Rule.of_terms ti si) $&$ lex gt eq ts ss)
;;

(* multiset cover *)
let mul_cover ss ts =
 let aux (ss,ts) =
  Made.mapi (fun j _ -> mcov_a_var (ss,ts,j)) ts >>= fun cs ->
  let m = List.length ss in
  let sm = Logic.big_and (List.map (fun ci -> ci <<> (constant m)) cs) in
  return (sm,cs)
 in get >>= fun c -> cache_m c.mcov_encodings aux (ss,ts)
;;

let mul gt eq ss ts =
 mul_cover ss ts >>= fun (cover,cs) ->
 let tcs = List.zip ts cs in
 mapi_and (fun i si -> mapi_and (fun j (tj,cj) -> 
  (return (cj <=> (constant i))) $->$ (gt (Rule.of_terms si tj))
  ) tcs) ss >>= fun phi ->
 return (cover <&> phi)
;;

let diff ?(cmp=compare) xs ys =
 let rec rem x = function
  | [] -> []
  | y::ys when cmp x y = 0 -> ys
  | y::ys -> y::(rem x ys)
 in List.foldl (flip rem) xs ys
;;


let mul_ge gt eq ss ts =
 let ss,ts = diff ~cmp:compare ss ts, diff ~cmp:compare ts ss in 
 if List.is_empty ss && List.is_empty ts then 
  return Logic.top
 else
  mul gt eq ss ts 
;;

let mul_gt gt eq ss ts = 
 let ss,ts = diff ~cmp:compare ss ts, diff ~cmp:compare ts ss in
 mul gt eq ss ts 
;;

let mul_eq gt eq ss ts =
 let ss,ts = diff ~cmp:compare ss ts, diff ~cmp:compare ts ss in 
 if List.is_empty ss && List.is_empty ts then 
  return Logic.top
 else 
  return Logic.bot
;;

(* to compute no_small_head and big_head multiset comparisons *)
let mul_g' cs p gt eq ss ts =
 let tcs = List.zip ts cs in
 let gt si tj = gt (Rule.of_terms si tj) in
 let has_val cj i = return (cj <=> (constant i)) in
 mapi_and (fun j (tj,cj) -> 
  (p tj) $->$ 
  (mapi_and (fun i si -> has_val cj i $->$ (gt si tj $&$ (p si))) ss)) 
 tcs
;;

let mul_ge' p gt eq ss ts =
 let ss,ts = diff ~cmp:compare ss ts, diff ~cmp:compare ts ss in
 mul_cover ss ts >>= fun (cover,cs) ->
 mul_g' cs p gt eq ss ts >>= fun phi ->
 return (cover <&> phi)
;;

let mul_gt' p gt eq ss ts =
 let ss,ts = diff ~cmp:compare ss ts, diff ~cmp:compare ts ss in
 mul_cover ss ts >>= fun (cover,cs) ->
 mul_g' cs p gt eq ss ts >>= fun phi ->
 map_or p ss >>= fun non_empty ->
 return (cover <&> phi <&> non_empty)
;;

(* emb_small candidates *)
let emb_sm f ts =
 let tf t = if Term.root t = Some f then Term.args t else [t] in 
 let rec emb hd =  function
  | [] -> []
  | ((Term.Var _) as t)::ts -> emb (hd@[t]) ts 
  | (Term.Fun(h,hs) as t)::ts when f=h -> emb (hd@[t]) ts
  | (Term.Fun(h,hs) as t)::ts -> 
   (List.map (fun hi -> (hd@(tf hi)@ts,h)) hs)@(emb (hd@[t]) ts)
 in List.map (Pair.apply (Term.make_fun f) id) (emb [] ts )
;;

(* #s > #t *)
let size_gt ss ts = 
 let (vs,fs),(vt,ft) = Pair.map (List.partition Term.is_var) (ss,ts) in
 let vs,vt = diff vs vt,diff vt vs in 
 (List.is_empty vt) && (List.length (vs@fs) > (List.length ft))
;;

(* #s = #t *)
let size_ge ss ts =
 let (vs,fs),(vt,ft) = Pair.map (List.partition Term.is_var) (ss,ts) in
 let vs,vt = diff vs vt,diff vt vs in
 (List.is_empty vt) && (List.length (vs@fs) >= (List.length ft))
;;

(* until quasi prec *)
let rpo_eq r = 
 let s,t = Rule.to_terms r in 
 return (if Term.compare s t = 0 then Logic.top else Logic.bot)
;;

let flat_rule r =
  Made.project flatten (Rule.to_terms r) >>= fun (l,r) ->
  return (Rule.of_terms l r)
;;

let rec rpo_gt rule =
 flat_rule rule >>= fun rule ->
(* Format.printf "check: %s > %s\n%!" (Term.to_string s) (Term.to_string t);*)
 let helper rule = 
  if is_embedded (Rule.invert rule) || not (Rule.is_rewrite_rule rule) then
   return Logic.bot
  else if is_proper_embedded rule then 
   return Logic.top
  else
   let s,t = Rule.to_terms rule in 
   match s,t with
   | Term.Var _, _ -> return Logic.bot
   | Term.Fun _, Term.Var x -> 
    return (if List.mem x (Term.vars s) then Logic.top else Logic.bot)
   | Term.Fun (f,ss), Term.Fun (g,ts) when f <> g ->
    map_or (fun si -> rpo_ge (Rule.of_terms si t)) ss  $|$                  (* (1) *)
    (prec_gt f g $&$ map_and (fun tj -> rpo_gt (Rule.of_terms s tj)) ts)      (* (2) *)
   | Term.Fun (f,ss), Term.Fun (_,ts)  ->
    is_ac f >>= fun f_is_ac ->
    if not f_is_ac then
     map_or (fun si -> rpo_ge (Rule.of_terms si t)) ss  $|$                 (* (1) *)
     ((is_lex f) $&$ (lex rpo_gt rpo_eq ss ts)) $|$                         (* (3) *)
     ((is_mul f) $&$ (mul_gt rpo_gt rpo_eq ss ts))                          (* (4) *)
    else (
     let gt s t = rpo_gt (Rule.of_terms s t) in
     let ge s t = rpo_ge (Rule.of_terms s t) in
     let cover h s' = prec_gt f h $&$ (ge s' t)  in
     map_or (fun (s',h) -> flatten s' >>= cover h) (emb_sm f ss) $|$        (* (5) *)
     (map_and (fun (t',h) -> prec_gt f h $->$ (gt s t')) (emb_sm f ts) $&$
      (no_small_head f ss ts)  $&$                                          (* (6) *)
      (if size_gt ss ts then return Logic.top                               (* (6b) *)
       else if not (size_ge ss ts) then big_head f ss ts                    (* (6a) *)
       else ((big_head f ss ts) $|$                                             
           (mul_gt rpo_gt rpo_eq ss ts))))                                  (* (6c) *)
)
  in
  get >>= fun c -> cache_m c.gt_encodings helper rule

and no_small_head f ss ts =
 let p u = match Term.root u with 
  None -> return Logic.top | Some h -> mnot (prec_gt f h) in
 mul_ge' p rpo_gt rpo_eq ss ts

and big_head f ss ts =
 if List.is_empty ss then return Logic.bot 
 else
  let p u = match Term.root u with    
   None -> return Logic.bot | Some h -> prec_gt h f in
  mul_gt' p rpo_gt rpo_eq ss ts

and rpo_ge rule = rpo_gt rule $|$ (rpo_eq rule) 
;;

(*
let case_five rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | (Term.Fun (f,ss)), Term.Fun (_,ts)  ->
    let ge s t = rpo_ge (Rule.of_terms s t) in
    map_or (fun (s',h) -> prec_gt f h $&$ (ge s' t)) (emb_sm f ss)
;;

let case_fivea rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | (Term.Fun (f,ss)), Term.Fun (_,ts)  ->
    let gt s t = rpo_gt (Rule.of_terms s t) in
    (fun (s',h) -> prec_gt f h $&$ (gt s' t)) (List.hd (emb_sm f ss))
;;

let case_fiveb rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | (Term.Fun (f,ss)), Term.Fun (_,ts)  ->
    let gt s t = rpo_gt (Rule.of_terms s t) in
    (fun (s',h) -> prec_gt f h $&$ (gt s' t)) (List.nth (emb_sm f ss) 1)
;;

let case_sixa rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | Term.Fun (f,ss), Term.Fun (_,ts)  ->
    let gt s t = rpo_gt (Rule.of_terms s t) in
    map_and (fun (t',h) -> prec_gt f h $->$ (gt s t')) (emb_sm f ts) $&$
    (no_small_head f ss ts)  $&$
    big_head f ss ts
;;

let no_small_head rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | Term.Fun (f,ss), Term.Fun (_,ts)  ->
    no_small_head f ss ts
;;

let big_head rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | Term.Fun (f,ss), Term.Fun (_,ts)  ->
    big_head f ss ts
;;

let case_sixb rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | Term.Fun (f,ss), Term.Fun (_,ts)  ->
    let gt s t = rpo_gt (Rule.of_terms s t) in
    map_and (fun (t',h) -> prec_gt f h $->$ (gt s t')) (emb_sm f ts) $&$
    (if size_gt ss ts then return Logic.top else return Logic.bot)
;;

let case_sixc rule =
 project flatten (Rule.to_terms rule) >>= fun (s,t) ->
 match s,t with
   | Term.Fun (f,ss), Term.Fun (_,ts)  ->
    let gt s t = rpo_gt (Rule.of_terms s t) in
    (*map_and (fun (t',h) -> prec_gt f h $->$ (gt s t')) (emb_sm f ts) $&$*)
    (if size_ge ss ts then mul_gt rpo_gt rpo_eq ss ts else return Logic.bot)
;;*)

let total_prec fs = 
 let comparable (f,g) = prec_gt f g $|$ prec_gt g f in
 map_and comparable (List.filter (fun (a,b) -> compare a b < 0) (List.square fs))
;;

let flat s w =
 let s, w = Pair.map Trs.to_list (s, w) in
 Made.project ((lift Trs.of_list) <.> (Made.map flat_rule)) (s,w)
;;

let encode_gt s = (if !(flags.direct) then map_and else map_or) rpo_gt s

let encode_ge s w = map_and rpo_ge (s@w)

let encode s w = get >>= fun c ->
 flat s w >>= fun (s,w) ->
 let fs = Trs.funs (Trs.union s w) in
 let s, w = Pair.map Trs.to_list (s, w) in
 encode_gt s $&$ encode_ge s w $&$ total_prec fs
;;

(* decode from assignment *)
let decode_rule ass rule = 
 get >>= fun c -> 
 flat_rule rule >>= fun rule ->
(* Format.printf "\nRule: %s\n%!" (Rule.to_string rule);
 eval_p (case_five rule) ass >>= fun b -> 
  Format.printf "(5) applies: %i \n%!" (if b then 1 else 0);
 eval_p (case_sixa rule) ass >>= fun b -> 
  Format.printf "(6a) applies: %i \n%!" (if b then 1 else 0);
 eval_p (no_small_head rule) ass >>= fun b ->
  Format.printf "no small head applies: %i \n%!" (if b then 1 else 0);
 eval_p (big_head rule) ass >>= fun b ->
  Format.printf "big head applies: %i \n%!" (if b then 1 else 0);
 eval_p (case_sixb rule) ass >>= fun b ->
  Format.printf "(6b) applies: %i \n%!" (if b then 1 else 0);
 eval_p (case_sixc rule) ass >>= fun b ->
  Format.printf "(6c) applies: %i \n%!" (if b then 1 else 0);*)
(* Logic.fprintf_assignment Format.std_formatter ass;*)
 lift not (eval_p (rpo_gt rule) ass)
;;

let decode_trs ass trs = 
 Made.filter (decode_rule ass) (Trs.to_list trs) >>= (return <.> Trs.of_list)
;;

let decode ass s w = get >>= fun c -> 
 lift2 Pair.make (decode_trs ass s) ((*decode_trs ass*) return w)
;;

let decode_status_f ass f = eval_p (is_mul f) ass >>= fun b -> return (f,b)

let decode_status ass s w =
 let fs = Trs.funs (Trs.union s w) in
 Made.sequence (List.map (decode_status_f ass) fs) >>= fun ps ->
 return (Status.of_list ps)
;;

let decode_prec_f ass f =
 eval_a (prec f) ass >>= fun p -> 
 return (f,int_of_string (Number.to_string p))
;;

let decode_prec ass s w =
 let fs = Trs.funs (Trs.union s w) in
 Made.sequence (List.map (decode_prec_f ass) fs) >>= fun ps ->
 return (Prec.of_list ps)
;;

let solve signature fs p = 
 let configurate s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 let (s,w) = P.get_sw p in
 let c = context signature (Trs.funs (Trs.union s w)) in
 Logic.run (
  Made.run c (encode s w >>= fun phi ->
  Made.liftm (Logic.solve ~solver:c.solver phi) >>= function
   | None -> return None
   | Some ass ->
    decode ass s w >>= fun (s',w') -> 
    decode_prec ass s w >>= fun prec -> 
    decode_status ass s w >>= fun status ->
    return (Some (make p (P.set_sw s' w' p) prec status))))
;;

(* wrap into state monad *)
let (>>=) = M.(>>=);;
let solve fs p = M.get >>= fun s -> M.return (solve s fs p);;
